% Last Updated: 21st Jul. 2011 
% Author: Ali Faisal
% Helsinki Institute for Information Technology
% Aalto University School of Science
% URL: http://users.ics.tkk.fi/faisal/  

% Last Updated (21st Jul 2011)
% LassoRegression v1.0

% Reference for citation:
% Faisal A et al., 2010 Inferring species interaction networks from species
% abundance data: A comparative evaluation of various statistical and
% machine learning methods, doi:10.1016/j.ecoinf.2010.06.005    

% INPUT
% data - an NxM matrix with N features/nodes/species/genes and M
% observations/samples. 
% 
% method (String)- Procedure used to set the sparsity parameter. 'direct'
% uses the pre-specified value, while 'xval' uses crossvalidation to
% determine the best optimum lambda between 1 and 100.     
% 
% partial_cor (String): If set to 'TRUE' it will transform output weights
% into an undirected networks using an estimator for partial correlations.
% (Default value 'FALSE' as this metric was not used in Faisal et al.,
% 2010)
%
% lambdaORk (Optional Integer): if method = 'direct' then lambdaORk
% represents the sparsity parameter, otherwise if the method is 'xval',
% lambdaORk can be used to set the no. of crossvalidation folds. (default
% value is 10-folds)   

% OUTPUT:
% weights - Edge strengths that are the weights of the predictor
% coefficients for each node. Basically  an NxN matrix where the zeros
% indicate  no connection between species/genes, and the non-zero values
% indicate the connection strength.      
%
% best_l - the value of (sparsity parameter) used.



% Thanks to Frank Dondelinger for sharing his insights on the code.
% Note: Function arrfit is taken from Grandvalet,98. The code is
% modification from the orignal (E.P. van Someren 2006) where it was
% applicable on time series data. It current version is extended to perform
% model selection and to suit stationary data.   

% For comments/feedback feel free to contact me at: "ali.faisal at aalto.fi"

function [weights, best_l] = gene_lasso(data, method, partial_cor, lambdaORk) 

% Number of observations
D = size(data, 2);

if nargin < 3
    partial_cor = 'FALSE';
end
if nargin < 4
  lambda = 1;
  k = 10;
elseif(strcmp(method , 'xval'))
  k = lambdaORk;
elseif(strcmp(method , 'direct'))
  lambda = lambdaORk;
else
   error('The second variable should be set to either direct Or xval') 
end

warning off
% Standardise data
data = standardizeCols(data')';

weights = zeros(size(data, 1), size(data, 1));

switch(method)
  % Use the lambda provided
  case 'direct'
    weights = estimateWeights(data, lambda);
    best_l = lambda;
    fprintf('\n Note: no model selection performed for the sparsity parameter\n\n')
  % Use cross-validation to find lambda
  case 'xval'
    lambda = 1:100;
    
    % Number of items in validation set
    val_set_num = floor(D / k);
      
    results = zeros(size(lambda, 2), k);
    
    for i = 1:k
      fprintf('Crossvalidation %g\n', i);
      
      % Split data into validation set and training set (note, split is
      % fixed, so the validation set is not a random sample)
      rest_data = data;
      val_set = data(:, (i-1)*val_set_num + 1:i*val_set_num);
      rest_data(:, (i-1)*val_set_num + 1:i*val_set_num) = [];
      
      for l = 1:size(lambda, 2)
        % Use LASSO
        weights = estimateWeights(rest_data, lambda(l));
      
        % Calculate mean squared error
        results(l, i) = mean(mean((weights'*val_set - val_set).^2));
      end
    end

    % Determine optimum lambda value
    [err, best_l] = min(mean(results, 2));
    
    % Use LASSO on whole data set with best lambda value
    weights = estimateWeights(data, lambda(best_l)); 
end

% weights(abs(weights) < 1e-4) = 0; Threshold for pruned edge.

if(strcmp(partial_cor, 'TRUE'))
    weights = sign(weights).*sqrt(abs(weights)'.*abs(weights));
end


function weights = estimateWeights(data, lambda)
% Function estimateWeights extracts the abundances for each species in
% turn, and uses lasso to determine the weights for predicting that species
% based on all the other ones.
%
% Inputs:
% data - An NxM matrix consisting of N species and M observations. Assumed
% to be already standardised.
% lambda - The parameter for restricting the weights
% 
% Output:
% weights - The weights found by the LASSO algorithm.
N = size(data, 1);
weights = zeros(N,N);

for n = 1:N
  %fprintf('Inferring species %g\n', n);

  Y = data(n, :);
  X = data;
  % Remove species N
  X(n, :) = [];

  % Invoke LASSO
  beta = arrfit(X', Y', lambda, 1e-2, 0);

  % Add weights to matrix
  weights(1:n-1, n) = beta(1:n-1);
  weights(n+1:end, n) = beta(n:end);

end


function [beta,msr,Stot] = arrfit(X,y,lambda,precision,bias)
%ARRFIT Adaptive Ridge Regression linear fit to data 
%       ARRFIT(X,y,lambda) finds the coefficients BETA, of 
%	the linear fit to the data, X(i,:)*BETA ~= y(i), 
%	minimizing the following expression:
%
%	sum((X*BETA-y).^2) + lambda * sum(abs(BETA))^2
%
%	[BETA,MSR] = ARRFIT(X,y,lambda,precision) returns the 
%	coefficients BETA and the mean squares residuals MSR.
%
%       X is the vector or matrix of input data, y is the vector
%	of output data. 
%	lambda (default=1) is a scalar or vector of penalization 
%	coefficients. If lambda is a vector, each column of BETA 
%	and MSR corresponds to the respective value of lambda. 
%	precision (default=1e-2) is an optional parameter of the 
%	procedure. It is a measure of the absolute and relative 
%	precisions required for BETA.  
 
%       22/06/98 Y. Grandvalet 
% 	Bibliography
%	@inproceedings{Grandvalet98a,
%	AUTHOR = 	"Grandvalet, Y.",
%	TITLE = 	"Least absolute shrinkage is equivalent to quadratic penalization ",
%	BOOKTITLE = 	"ICANN 98",
%	EDITOR = 	"Niklasson, L. and Bod{\'e}n, M. and Ziemske, T.",
%	VOLUME = 	"1",
%	PAGES = 	"201--206",
%	PUBLISHER = 	"Springer",
%	SERIES = 	"Perspectives in Neural Computing",
%	YEAR = 		"1998"}

if nargin<5,
    bias = 0;
if nargin < 4;
   precision = 1e-2;
   if nargin < 3;
      lambda = 1;
      if nargin < 2;
    	error('ARRFIT requires at least two input arguments.');
      end;
   end;
end;
end;
precision = precision.^2;
 
% Check that matrix (X) and vector (y) have compatible dimensions
 
[n,d]   = size(X);
[ny,dy] = size(y);
if ny~=n, 
    error('The number of rows in y must equal the number of rows in X.'); 
end 
if dy ~= 1, 
    error('y must be a vector, not a matrix'); 
end

% Check that (lambda) has correct dimensions
 
[nl,dl] = size(lambda);
if dl ~= 1 & nl ~= 1, 
    error('lambda must be a scalar or vector.');
end 
[nl] = max([nl,dl]);

% Check that (precision) has correct dimensions
 
if length(precision) ~= 1, 
    error('precision must be a scalar.');
end 

% Initializations

beta = zeros(d,nl); 

XX = (X'*X);
Xy = (X'*y);

for i=1:nl
    if	lambda(i)==Inf;
	beta(:,i) = zeros(d,1);
    else;
        if bias
    	    Lambda = [lambda(i)*ones(d-1,1);0];
        else
          	Lambda = lambda(i)*ones(d,1);
        end;
	    U = chol(XX + diag(Lambda));
    	betanew = U\(U'\Xy);
    	stop    = 0;
        iter = 0;
    	while (~stop);
             betaold      = betanew;
	         normbetaold  = abs(betaold)./mean(abs(betaold));
             ind          = find( normbetaold > precision );
             Lambda(ind)  = (d*lambda(i))./normbetaold(ind);
             if bias,
                Lambda(end) = 0;
             end;
	         betanew      = zeros(d,1); 
	         U            = chol(XX(ind,ind) + diag(Lambda(ind)));
    	     betanew(ind) = U\(U'\Xy(ind));
             stop         = max( abs(betaold-betanew)./(1+abs(betanew)) ) < precision;
             iter = iter+1;
             if iter>10000,
                 errordlg('Maximum iterations exceeded','No convergence');
                 return;
             end;
    	end
    	beta(:,i) = betanew;
    end;
end;

if nargout > 1
   msr = sum( (X*beta - y(:,ones(nl,1))).^2 )/n;
end;
if nargout > 2
   Stot = sum( (X*beta - y(:,ones(nl,1))).^2 )/n + lambda * sum(abs(beta))^2;
end;

return;

function [S, mu, sigma2] = standardizeCols(M, mu, sigma2)
% Make each column of M be zero mean, std 1.
%
% If mu, sigma2 are omitted, they are computed from M

[nrows ncols] = size(M);

M = double(M);
if nargin < 2
  mu = mean(M);
  sigma2 = std(M);
  sigma2(sigma2 < eps) = 1;
end

S = M - repmat(mu, [nrows 1]);
if ncols > 0
S = S ./ repmat(sigma2, [nrows 1]);
end
